from django.contrib.auth import get_user_model
from django.db import models
from django.utils.translation import gettext_lazy as _

User = get_user_model()


class CASAccount(models.Model):
    """Information about CAS accounts.

    A user is given an instance of this model iff they have a CAS account.

    Instances of this model should only be created by the `ENSCASBackend` authentication
    backend.
    """

    user = models.OneToOneField(
        User,
        verbose_name=_("utilisateurice"),
        on_delete=models.CASCADE,
        related_name="cas_account",
    )
    cas_login = models.CharField(
        verbose_name=_("login CAS"),
        max_length=1023,
        blank=False,
        unique=True,
    )
    entrance_year = models.SmallIntegerField(
        verbose_name=_("année de création du compte CAS"), blank=False, null=False
    )

    class Meta:
        verbose_name = _("Compte CAS")
        verbose_name_plural = _("Comptes CAS")

    def __str__(self):
        return _(
            "compte CAS %(cas_login)s (promo %(entrance_year)s) lié à %(user)s"
        ) % {
            "cas_login": self.cas_login,
            "entrance_year": self.entrance_year,
            "user": self.user.username,
        }


class OldCASAccount(models.Model):
    """Information about expired CAS accounts

    A user is given an instance of this model iff they had a CAS account that expired.

    Instances of this model should not be created with a new user.
    """

    user = models.OneToOneField(
        User,
        verbose_name=_("utilisateurice"),
        on_delete=models.CASCADE,
        related_name="old_cas_account",
    )

    cas_login = models.CharField(
        verbose_name=_("ancien login CAS"), max_length=1023, blank=False
    )

    entrance_year = models.SmallIntegerField(
        verbose_name=_("année de création du compte CAS"), blank=False, null=False
    )

    class Meta:
        # `unique_together` to be deprecated soon : we use `constraints`
        constraints = [
            models.UniqueConstraint(
                fields=["cas_login", "entrance_year"],
                name="clipper_year_uniqueness",
            )
        ]
        verbose_name = _("Ancien compte CAS")
        verbose_name_plural = _("Anciens comptes CAS")

    def __str__(self):
        return _(
            "Ancien compte CAS %(cas_login)s (promo %(entrance_year)s) lié à %(user)s"
        ) % {
            "cas_login": self.cas_login,
            "entrance_year": self.entrance_year,
            "user": self.user.username,
        }



"""Internal utility functions used by authens."""
from urllib.parse import urlunparse

from cas import CASClient


def get_cas_client(request):
    """Return a CAS client configured for SPI's CAS."""
    return CASClient(
        version=3,
        service_url=urlunparse(
            (request.scheme, request.get_host(), request.path, "", "", "")
        ),
        server_url="https://cas.eleves.ens.fr/",
    )


def parse_entrance_year(home_dir):
    """Infer the entrance year of a CAS account from their home directory."""

    # The home directory of a user is of the form /users/YEAR/DEPARTMENT/CAS_LOGIN where
    # YEAR is a 2-digit number representing the entrance year of the student. We get the
    # entrance year from there.

    if home_dir is None:
        return None

    dirs = home_dir.split("/")
    if len(dirs) < 3 or not dirs[2].isdecimal() or dirs[1] != "users":
        raise ValueError("Invalid home directory: {}".format(home_dir))

    # Expand the 2-digit entrance year into 4 digits.
    # This will break in 2080.
    year = int(dirs[2])
    if year >= 80:
        return 1900 + year
    else:
        return 2000 + year



from unittest import mock

from django.contrib.auth import authenticate, get_user_model
from django.test import TestCase

from authens.backends import ENSCASBackend
from authens.models import CASAccount, OldCASAccount
from authens.tests.cas_utils import FakeCASClient

UserModel = get_user_model()


class TestCASBackend(TestCase):
    def test_usernames_uniqueness(self):
        UserModel.objects.create(username="toto")
        UserModel.objects.create(username="toto2")
        backend = ENSCASBackend()
        for _ in range(10):
            username = backend.get_free_username("toto", None)
            self.assertFalse(UserModel.objects.filter(username=username).exists())
            UserModel.objects.create(username=username)
        for _ in range(10):
            username = backend.get_free_username("tutu", None)
            self.assertFalse(UserModel.objects.filter(username=username).exists())
            UserModel.objects.create(username=username)

    def test_email(self):
        backend = ENSCASBackend()
        attributes = {
            "email": "toto@example.com",
            "homeDirectory": "/users/19/info/toto",
        }
        user = backend._get_or_create("toto", attributes)
        self.assertEqual(user.email, "toto@example.com")

    @mock.patch("authens.backends.get_cas_client")
    def test_cas_user_creation(self, mock_cas_client):
        # Make `get_cas_client` return a dummy CAS client for testing purpose.
        fake_cas_client = FakeCASClient()
        mock_cas_client.return_value = fake_cas_client

        # First authentication
        user1 = authenticate(None, ticket="dummy ticket")
        # Should create a user and a Clipper instance.
        self.assertEqual(CASAccount.objects.count(), 1)
        self.assertEqual(UserModel.objects.count(), 1)
        clipper1 = CASAccount.objects.get()
        self.assertEqual(clipper1.user, user1)
        self.assertEqual(clipper1.cas_login, fake_cas_client.cas_login)
        self.assertEqual(clipper1.entrance_year, fake_cas_client.entrance_year)

        # Second time we authenticate the same user.
        user2 = authenticate(None, ticket="dummy ticket")
        # Should return the same user and should not create anything.
        self.assertEqual(CASAccount.objects.count(), 1)
        self.assertEqual(UserModel.objects.count(), 1)
        clipper2 = CASAccount.objects.get()
        self.assertEqual(clipper1, clipper2)
        self.assertEqual(user1, user2)
        self.assertEqual(user2, clipper2.user)

    @mock.patch("authens.backends.get_cas_client")
    def test_name_conflict_handling(self, mock_cas_client):
        # Make `get_cas_client` return a dummy CAS client for testing purpose.
        fake_cas_client = FakeCASClient()
        mock_cas_client.return_value = fake_cas_client

        # Create a regular user (without a CAS account) with the same username as the
        # CAS user before her first login.
        regular_user = UserModel.objects.create_user(username=fake_cas_client.cas_login)

        # First login of the CAS user.
        cas_user = authenticate(None, ticket="dummy ticket")

        self.assertNotEqual(regular_user, cas_user)

    @mock.patch("authens.backends.get_cas_client")
    def test_cas_conflict_handling(self, mock_cas_client):
        # Make `get_cas_client` return a dummy CAS client for testing purpose.
        fake_cas_client = FakeCASClient()
        mock_cas_client.return_value = fake_cas_client

        # Create an old CAS user with the same cas_id as the new CAS user before her
        # first login. This tests the behaviour of the library in the (unlikely)
        # scenario where the SPI decides to give a clipper login that used to be someone
        # else's to a new user.
        old_user = UserModel.objects.create_user(username=fake_cas_client.cas_login)
        old_clipper = CASAccount.objects.create(
            user=old_user,
            cas_login=fake_cas_client.cas_login,
            entrance_year=fake_cas_client.entrance_year - 10,
        )

        # Log the new 'johndoe' in.
        new_user = authenticate(None, ticket="dummy ticket")
        new_clipper = new_user.cas_account

        # Check that it gets a fresh user and a fresh clipper account
        self.assertNotEqual(old_user, new_user)
        self.assertNotEqual(old_clipper, new_clipper)

        # Check that the created CAS account matches the old one
        self.assertEqual(new_clipper.cas_login, fake_cas_client.cas_login)
        self.assertEqual(new_clipper.entrance_year, fake_cas_client.entrance_year)

        # Check deprecation of the old CAS account
        old_user.refresh_from_db()
        self.assertFalse(hasattr(old_user, "cas_account"))
        self.assertTrue(hasattr(old_user, "old_cas_account"))
        old_cas = old_user.old_cas_account
        self.assertEqual(old_cas.cas_login, old_clipper.cas_login)
        self.assertEqual(old_cas.entrance_year, old_clipper.entrance_year)


class TestOldCASBackend(TestCase):
    def test_simple_auth(self):
        user = UserModel.objects.create_user(username="johndoe31", password="password")
        # Decoy user that may be authenticated by mistake
        UserModel.objects.create_user(username="johndoe", password="password")
        OldCASAccount.objects.create(user=user, cas_login="johndoe", entrance_year=2014)

        auth_user = authenticate(
            None, cas_login="johndoe", entrance_year=2014, password="password"
        )
        self.assertEqual(auth_user, user)



# Generated by Django 3.0.6 on 2020-05-17 11:58
import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="CASAccount",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "cas_login",
                    models.CharField(
                        max_length=1023, unique=True, verbose_name="login CAS"
                    ),
                ),
                (
                    "entrance_year",
                    models.SmallIntegerField(
                        verbose_name="année de création du compte CAS"
                    ),
                ),
                (
                    "user",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="cas_account",
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="utilisateurice",
                    ),
                ),
            ],
            options={
                "verbose_name": "Compte CAS",
                "verbose_name_plural": "Comptes CAS",
            },
        ),
    ]



from django.contrib.auth import get_user_model
from django.db import transaction

from authens.models import CASAccount, OldCASAccount
from authens.signals import post_cas_connect
from authens.utils import get_cas_client, parse_entrance_year

UserModel = get_user_model()


class ENSCASError(Exception):
    pass


class ENSCASBackend:
    """AuthENS CAS authentication backend.

    Implement standard CAS v3 authentication and handles username clashes with non-CAS
    accounts and potential old CAS accounts.

    Every user connecting via CAS is given an `authens.models.CASAccount` instance which
    remembers her CAS login and her entrance year (the year her CAS account was
    created).
    At each connection, we search for a CAS account with the given CAS login and create
    one if none exists. In case the CAS account's entrance year does not match the
    entrance year given by CAS, it means it is a old account and it must be deleted. The
    corresponding user can still connect using regular Django authentication.
    """

    def authenticate(self, request, ticket=None):
        cas_client = get_cas_client(request)
        cas_login, attributes, _ = cas_client.verify_ticket(ticket)

        if cas_login is None:
            # Authentication failed
            return None
        cas_login = self.clean_cas_login(cas_login)

        if request:
            request.session["CASCONNECTED"] = True

        return self._get_or_create(cas_login, attributes)

    def clean_cas_login(self, cas_login):
        return cas_login.strip().lower()

    def get_free_username(self, cas_login, attributes):
        """Find an available username for a new user.

        If you override this method, make sure it returns a username that is not taken
        by any existing user.
        This method will be called from within an atomic database transation (don't
        worry about data races).
        """

        # Ideally we can have username = cas_login.
        if not UserModel.objects.filter(username=cas_login).exists():
            return cas_login

        # Otherwise, add a numerical suffix.

        # If there was a numerical suffix already, remove it:
        # https://git.eleves.ens.fr/klub-dev-ens/authens/issues/5
        i = len(cas_login) - 1
        while i >= 0 and cas_login[i] in "0123456789":
            i -= 1
        radical = cas_login[: i + 1]

        # Find an integer i such that radical + str(i) is not taken.
        taken = UserModel.objects.values_list("username", flat=True)
        taken = taken.filter(username__startswith=radical)
        i = 2
        while radical + str(i) in taken:
            i += 1
        return radical + str(i)

    def create_user(self, username, attributes):
        """Create a CAS user, base method that can be overrided to add more
        information.
        """

        email = attributes.get("email")
        return UserModel.objects.create_user(username=username, email=email)

    def _get_or_create(self, cas_login, attributes):
        """Handles account retrieval, creation and invalidation as described above.

        - If no CAS account exists, create one;
        - If a CAS account exists, but with the wrong entrance year, convert it to
        an OldCASAccount instance, and create a fresh CAS Account with the correct year.
        - If a matching CAS account exists, retrieve it.
        """

        entrance_year = parse_entrance_year(attributes.get("homeDirectory"))
        if entrance_year is None:
            raise ENSCASError("Entrance year not available")

        account_created = False
        with transaction.atomic():
            try:
                user = UserModel.objects.get(cas_account__cas_login=cas_login)
                if user.cas_account.entrance_year != entrance_year:
                    OldCASAccount.objects.create(
                        user=user,
                        entrance_year=user.cas_account.entrance_year,
                        cas_login=cas_login,
                    )
                    user.cas_account.delete()
                    user = None
            except UserModel.DoesNotExist:
                user = None

            if user is None:
                username = self.get_free_username(cas_login, attributes)
                user = self.create_user(username, attributes)
                CASAccount.objects.create(
                    user=user, entrance_year=entrance_year, cas_login=cas_login
                )
                account_created = True

        post_cas_connect.send(
            UserModel,
            instance=user,
            created=account_created,
            cas_login=cas_login,
            attributes=attributes,
        )
        return user

    # Django boilerplate.
    def get_user(self, user_id):
        try:
            return UserModel.objects.get(pk=user_id)
        except UserModel.DoesNotExist:
            return None


class OldCASBackend:
    """Authentication backend for old CAS accounts.

    Given a CAS login, an entrance year and a password, first finds the matching
    OldCASAccount instance (if it exists), then checks the given password with
    the user associated to this account.
    """

    def authenticate(self, request, cas_login=None, password=None, entrance_year=None):
        if cas_login is None or password is None or entrance_year is None:
            return

        try:
            old_cas_acc = OldCASAccount.objects.get(
                cas_login=cas_login, entrance_year=entrance_year
            )
            user = old_cas_acc.user
        except OldCASAccount.DoesNotExist:
            # As in Django's ModelBackend, we run the password hasher once
            # to mitigate timing attacks
            UserModel().set_password(password)
        else:
            if user.check_password(password) and self.user_can_authenticate(user):
                return user

    def user_can_authenticate(self, user):
        # Taken from Django's ModelBackend
        is_active = getattr(user, "is_active", None)
        return is_active or is_active is None

    # Django boilerplate.
    def get_user(self, user_id):
        try:
            return UserModel.objects.get(pk=user_id)
        except UserModel.DoesNotExist:
            return None



import setuptools

with open("README.md", "r") as file:
    long_description = file.read()


setuptools.setup(
    name="authens",
    version="0.1b1",
    author="Klub Dev ENS",
    author_email="klub-dev@ens.fr",
    description="CAS Authentication at the ENS",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://git.eleves.ens.fr/klub-dev-ens/authens",
    packages=setuptools.find_packages(),
    include_package_data=True,
    classifiers=[
        "Environment :: Web Environment",
        "Framework :: Django",
        "Framework :: Django :: 2.2",
        "Framework :: Django :: 3.0",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
    ],
    python_requires=">=3.5",
    install_requires=["Django>=2.2", "python-ldap>=3,<4", "python-cas>=1.5,<2"],
)



"""Helper functions to get CAS metadata and create CAS accounts."""

# TODO: make the python-ldap dependency optional
import ldap
from django.conf import settings
from django.contrib.auth import get_user_model

from authens import conf as default_conf
from authens.models import CASAccount, OldCASAccount
from authens.utils import parse_entrance_year

User = get_user_model()


def _extract_ldap_info(entry, field):
    dn, attrs = entry
    return attrs[field][0].decode("utf-8")


def fetch_cas_account(cas_login):
    """Issue an LDAP connection to retrieve metadata associated to a CAS account."""

    # Don't trust the user! Only accept alphanumeric account names.
    if not cas_login.isalnum():
        raise ValueError("Illegal CAS login: {}".format(cas_login))

    ldap_url = getattr(settings, "LDAP_SERVER_URL", default_conf.LDAP_SERVER_URL)
    ldap_obj = ldap.initialize(ldap_url)
    res = ldap_obj.search_s(
        "dc=spi,dc=ens,dc=fr",
        ldap.SCOPE_SUBTREE,
        "(uid={})".format(cas_login),
        ["uid", "cn", "homeDirectory"],
    )
    if not res:
        return None

    if len(res) != 1:
        raise RuntimeError("LDAP returned too many results: {}".format(res))
    (res,) = res

    assert _extract_ldap_info(res, "uid") == cas_login
    return {
        "cn": _extract_ldap_info(res, "cn"),
        "entrance_year": parse_entrance_year(_extract_ldap_info(res, "homeDirectory")),
    }


def register_cas_account(user: User, cas_login: str) -> CASAccount:
    """Register a user as a CAS user and return the newly created CASAccount."""

    if not cas_login:
        raise ValueError("cas_login must be non-empty")
    if CASAccount.objects.filter(cas_login=cas_login).exists():
        raise ValueError("A CAS account named '{}' exists already".format(cas_login))
    if CASAccount.objects.filter(user=user).exists():
        raise ValueError("User '{}' already has a CAS account".format(user))
    if OldCASAccount.objects.filter(user=user).exists():
        raise ValueError("User '{}' has an old CAS account".format(user))

    ldap_info = fetch_cas_account(cas_login)
    if ldap_info is None:
        raise ValueError("There is no LDAP user for id '{}'".format(cas_login))

    entrance_year = ldap_info["entrance_year"]
    return CASAccount.objects.create(
        user=user, cas_login=cas_login, entrance_year=entrance_year
    )



from django.conf import settings
from django.urls import path

from authens import conf as default_conf, views

app_name = "authens"
urlpatterns = [
    path("login/choose", views.LoginSwitchView.as_view(), name="login"),
    path("login/cas", views.CASLoginView.as_view(), name="login.cas"),
    path("login/pwd", views.PasswordLoginView.as_view(), name="login.pwd"),
    path("logout", views.LogoutView.as_view(), name="logout"),
    path("reset/pwd", views.PasswordResetView.as_view(), name="reset.pwd"),
    path(
        "reset/pwd/<uidb64>/<token>/",
        views.PasswordResetConfirmView.as_view(),
        name="reset.pwd.confirm",
    ),
]

if getattr(settings, "AUTHENS_USE_OLDCAS", default_conf.AUTHENS_USE_OLDCAS):
    urlpatterns += (
        path("login/oldcas", views.OldCASLoginView.as_view(), name="login.oldcas"),
    )
